// $Id: DualRand.cpp,v 1.1.1.1 2000/04/03 21:03:44 wilko Exp $
// -*- C++ -*-
//
// -----------------------------------------------------------------------
//                           Hep Random
//                        --- DualRand ---
//                   class implementation file
// -----------------------------------------------------------------------
//  Exclusive or of a feedback shift register and integer congruence
//  random number generator.  The feedback shift register uses offsets
//  127 and 97.  The integer congruence generator uses a different
//  multiplier for each stream.  The multipliers are chosen to give
//  full period and maximum "potency" for modulo 2^32.  The period of
//  the combined random number generator is 2^159 - 2^32, and the
//  sequences are different for each stream (not just started in a
//  different place).
//
//  In the original generator used on ACPMAPS:
//  The feedback shift register generates 24 bits at a time, and
//  the high order 24 bits of the integer congruence generator are
//  used.
//
//  Instead, to conform with more modern engine concepts, we generate
//  32 bits at a time and use the full 32 bits of the congruence
//  generator.
//
//  References:
//	Knuth
//	Tausworthe
//	Golomb
//=========================================================================
// Ken Smith      - Removed pow() from flat() method:           21 Jul 1998
//                - Added conversion operators:                  6 Aug 1998
// J. Marraffino  - Added some explicit casts to deal with
//                  machines where sizeof(int) != sizeof(long)  22 Aug 1998
//=========================================================================

#include "CLHEP/Random/DualRand.h"
//#include "CLHEP/String/Strings.h"
#include <iomanip>

// Number of instances with automatic seed selection
HepInt DualRand::numEngines = 0;

// The following constructors (excluding the istream constructor)  fill
// the bits of the tausworthe and the starting state of the integer
// congruence based on the seed.  In addition, it sets up the multiplier
// for the integer congruence based on the stream number, so you have 
// absolutely independent streams.

DualRand::DualRand() 
: mantissa_bit_32 ( pow(0.5,32. ) ),
  mantissa_bit_53 ( pow(0.5,53. ) ),
  mantissa_bit_54 ( pow(0.5,54. ) ),
  mantissa_bit_106( pow(0.5,106.) ),
  tausworthe (1234567 + numEngines + 175321),
  integerCong(69607 * tausworthe + 54329, numEngines)
{
  theSeed = 1234567;
  ++numEngines;
}

DualRand::DualRand(long seed)
: mantissa_bit_32 ( pow(0.5,32. ) ),
  mantissa_bit_53 ( pow(0.5,53. ) ),
  mantissa_bit_54 ( pow(0.5,54. ) ),
  mantissa_bit_106( pow(0.5,106.) ),
  tausworthe ((unsigned int)seed + numEngines + 175321),
  integerCong(69607 * tausworthe + 54329, numEngines)
{
  theSeed = seed;
}

DualRand::DualRand(std::istream & is) 
: mantissa_bit_32 ( pow(0.5,32. ) ),
  mantissa_bit_53 ( pow(0.5,53. ) ),
  mantissa_bit_54 ( pow(0.5,54. ) ),
  mantissa_bit_106( pow(0.5,106.) )
{
  is >> *this;
}

DualRand::DualRand(HepInt rowIndex, HepInt colIndex)
: mantissa_bit_32 ( pow(0.5,32. ) ),
  mantissa_bit_53 ( pow(0.5,53. ) ),
  mantissa_bit_54 ( pow(0.5,54. ) ),
  mantissa_bit_106( pow(0.5,106.) ),
  tausworthe (rowIndex + numEngines * colIndex + 175321),
  integerCong(69607 * tausworthe + 54329, numEngines)
{
  theSeed = rowIndex;
}

DualRand::~DualRand() { }

DualRand::DualRand( const DualRand & p )
: mantissa_bit_32 ( pow(0.5,32. ) ),
  mantissa_bit_53 ( pow(0.5,53. ) ),
  mantissa_bit_54 ( pow(0.5,54. ) ),
  mantissa_bit_106( pow(0.5,106.) )
{
  *this = p;
}

DualRand & DualRand::operator=( const DualRand & p ) {
  if( this != &p ) {
     tausworthe  = p.tausworthe;
     integerCong = p.integerCong;
  }
  return *this;
}

HepDouble DualRand::flat() {
  unsigned int ic ( integerCong );
  unsigned int t  ( tausworthe  );
  return ( (t  ^ ic) * mantissa_bit_32 +              // most significant part
           (t >> 11) * mantissa_bit_53 +              // fill in remaining bits
           mantissa_bit_54 - mantissa_bit_106         // make sure non-zero
         );
}

void DualRand::flatArray(const HepInt size, HepDouble* vect) {
  for (HepInt i = 0; i < size; ++i) {
    vect[i] = flat();
  }
}

void DualRand::setSeed(long seed, HepInt) {
  theSeed = seed;
  tausworthe  = Tausworthe((unsigned int)seed + numEngines + 175321);
  integerCong = IntegerCong(69607 * tausworthe + 54329, numEngines);
}

void DualRand::setSeeds(const long * seeds, HepInt) {
  setSeed(seeds ? *seeds : 1234567, 0);
  theSeeds = seeds;
}

void DualRand::saveStatus(const char filename[]) const {
  std::ofstream outFile(filename, std::ios::out);
  if (!outFile.bad()) {
    outFile << std::setprecision(20) << theSeed << std::endl;
    tausworthe.put(outFile);
    integerCong.put(outFile);
  }
}

void DualRand::restoreStatus(const char filename[]) {
  std::ifstream inFile(filename, std::ios::in);
  if (!inFile.bad()) {
    inFile >> theSeed;
    tausworthe.get(inFile);
    integerCong.get(inFile);
  }
}

void DualRand::showStatus() const {
  std::cout << std::setprecision(20) << std::endl;
  std::cout <<         "-------- DualRand engine status ---------"
	    << std::endl;
  std::cout << "Initial seed          = " << theSeed << std::endl;
  std::cout << "Tausworthe generator  = " << std::endl;
  tausworthe.put(std::cout);
  std::cout << "\nIntegerCong generator = " << std::endl;
  integerCong.put(std::cout);
  std::cout << std::endl << "-----------------------------------------"
	    << std::endl;
}

DualRand::operator float() {
  return (HepFloat) (( integerCong ^ tausworthe ) * mantissa_bit_32 );
}

DualRand::operator unsigned int() {
  return (integerCong ^ tausworthe) & 0xffffffff;
}
/*
std::ostream & operator<< (std::ostream & os, const DualRand & e) {
  HepString beginMarker = "DualRand-begin";
  HepString endMarker   = "DualRand-end";

  os << " " << beginMarker << " ";
  os << std::setprecision(20) << e.theSeed << " ";
  e.tausworthe.put(os);
  e.integerCong.put(os);
  os << " " <<  endMarker  << " ";
  return os;
}
*/
/*
std::istream & operator>> (std::istream & is, DualRand & e) {
  HepString beginMarker = " ";
  HepString endMarker   = " ";

  is >> std::ws >> beginMarker;
  if (beginMarker != "DualRand-begin") {
    is.clear(std::ios::badbit | is.rdstate());
    std::cerr << "\nInput mispositioned or"
	      << "\nDualRand state description missing or"
	      << "\nwrong engine type found." << std::endl;
    return is;
  }
  is >> e.theSeed;
  e.tausworthe.get(is);
  e.integerCong.get(is);
  is >> std::ws >> endMarker;
  if (endMarker != "DualRand-end") {
    is.clear(std::ios::badbit | is.rdstate());
    std::cerr << "\nDualRand state description incomplete."
	      << "\nInput stream is probably mispositioned now." << std::endl;
    return is;
  }
  return is;
}
*/
DualRand::Tausworthe::Tausworthe() {
  words[0] = 1234567;
  for (wordIndex = 1; wordIndex < 4; ++wordIndex) {
    words[wordIndex] = 69607 * words[wordIndex-1] + 54329;
  } 
}

DualRand::Tausworthe::Tausworthe(unsigned int seed) {
  words[0] = seed;
  for (wordIndex = 1; wordIndex < 4; ++wordIndex) {
    words[wordIndex] = 69607 * words[wordIndex-1] + 54329;
  }
}

DualRand::Tausworthe::operator unsigned int() {

// Mathematically:  Consider a sequence of bits b[n].  Repeatedly form
// b[0]' = b[127] ^ b[97]; b[n]' = b[n-1].  This sequence will have a very
// long period (2**127-1 according to Tausworthe's work).

// The actual method used relies on the fact that the operations needed to
// form bit 0 for up to 96 iterations never depend on the results of the
// previous ones.  So you can actually compute many bits at once.  In fact
// you can compute 32 at once -- despite 127 - 97 < 32 -- but 24 was used in
// the method used in Canopy, where they wanted only single-precision float
// randoms.  I will do 32 here.

// When you do it this way, this looks disturbingly like the dread lagged XOR
// Fibonacci.  And indeed, it is a lagged Fibonacii, F(4,3, op) with the op
// being the XOR of a combination of shifts of the two numbers.  Although
// Tausworthe asserted excellent properties, I would be scared to death.
// However, the shifting and bit swapping really does randomize this in a
// serious way.

// Statements have been made to the effect that shift register sequences fail
// the parking lot test because they achieve randomness by multiple foldings,
// and this produces a characteristic pattern.  We observe that in this
// specific algorithm, which has a fairly long lever arm, the foldings become
// effectively random.  This is evidenced by the fact that the generator
// does pass the Diehard tests, including the parking lot test.

// To avoid shuffling of variables in memory, you either have to use circular
// pointers (and those give you ifs, which are also costly) or compute at least
// a few iterations at once.  We do the latter.  Although there is a possible
// trade of room for more speed, by computing and saving 256 instead of 128
// bits at once, I will stop at this level of optimization.

// To remind:  Each (32-bit) step takes the XOR of bits [127-96] with bits
// [95-64] and places it in bits [0-31].  But in the first step, we designate
// word0 as bits [0-31], in the second step, word 1 (since the bits it holds
// will no longer be needed), then word 2, then word 3.  After this, the
// stream contains 128 random bits which we will use as 4 valid 32-bit
// random numbers.

// Thus at the start of the first step, word[0] contains the newest (used)
// 32-bit random, and word[3] the oldest.   After four steps, word[0] again
// contains the newest (now unused) random, and word[3] the oldest.
// Bit 0 of word[0] is logically the newest bit, and bit 31 of word[3]
// the oldest.

  if (wordIndex <= 0) {
    for (wordIndex = 0; wordIndex < 4; ++wordIndex) {
      words[wordIndex] = ( (words[(wordIndex+1) & 3] << 1 ) |
                                   (words[wordIndex] >> 31)   )
                       ^ ( (words[(wordIndex+1) & 3] << 31) |
                                   (words[wordIndex] >>  1)   );
    }
  }
  return words[--wordIndex] & 0xffffffff;
}
/*
void DualRand::Tausworthe::put(std::ostream & os) const {
  HepString beginMarker = "Tausworthe-begin";
  HepString endMarker   = "Tausworthe-end";

  os << " " << beginMarker << " ";
  os << std::setprecision(20);
  for (HepInt i = 0; i < 4; ++i) {
    os << words[i] << " ";
  }
  os << wordIndex;
  os << " " <<  endMarker  << " ";
}
*/
/*
void DualRand::Tausworthe::get(std::istream & is) {
  HepString beginMarker = " ";
  HepString endMarker   = " ";

  is >> std::ws >> beginMarker;
  if (beginMarker != "Tausworthe-begin") {
    is.clear(std::ios::badbit | is.rdstate());
    std::cerr << "\nInput mispositioned or"
	      << "\nTausworthe state description missing or"
	      << "\nwrong engine type found." << std::endl;
  }
  for (HepInt i = 0; i < 4; ++i) {
    is >> words[i];
  }
  is >> wordIndex;
  is >> std::ws >> endMarker;
  if (endMarker != "Tausworthe-end") {
    is.clear(std::ios::badbit | is.rdstate());
    std::cerr << "\nTausworthe state description incomplete."
	      << "\nInput stream is probably mispositioned now." << std::endl;
  }
}
*/
DualRand::IntegerCong::IntegerCong() 
: state((unsigned int)3758656018),
  multiplier(66565),
  addend(12341) 
{
}

DualRand::IntegerCong::IntegerCong(unsigned int seed, HepInt streamNumber)
: state(seed),
  multiplier(65536 + 1024 + 5 + (8 * 1017 * streamNumber)),
  addend(12341)
{
  // As to the multiplier, the following comment was made:
  // We want our multipliers larger than 2^16, and equal to
  // 1 mod 4 (for max. period), but not equal to 1 mod 8 
  // (for max. potency -- the smaller and higher dimension the 
  // stripes, the better)

  // All of these will have fairly long periods.  Depending on the choice
  // of stream number, some of these will be quite decent when considered
  // as independent random engines, while others will be poor.  Thus these
  // should not be used as stand-alone engines; but when combined with a
  // generator known to be good, they allow creation of millions of good
  // independent streams, without fear of two streams accidentally hitting
  // nearby places in the good random sequence.
}

DualRand::IntegerCong::operator unsigned int() {
  return state = (state * multiplier + addend) & 0xffffffff;
}
/*
void DualRand::IntegerCong::put(std::ostream & os) const {
  HepString beginMarker = "IntegerCong-begin";
  HepString endMarker   = "IntegerCong-end";

  os << " " << beginMarker << " ";
  os << std::setprecision(20);
  os << state << " " << multiplier << " " << addend;
  os << " " <<  endMarker  << " ";
}
*/
/*
void DualRand::IntegerCong::get(std::istream & is) {
  HepString beginMarker = " ";
  HepString endMarker   = " ";

  is >> std::ws >> beginMarker;
  if (beginMarker != "IntegerCong-begin") {
    is.clear(std::ios::badbit | is.rdstate());
    std::cerr << "\nInput mispositioned or"
	      << "\nIntegerCong state description missing or"
	      << "\nwrong engine type found." << std::endl;
  }
  is >> state >> multiplier >> addend;
  is >> std::ws >> endMarker;
  if (endMarker != "IntegerCong-end") {
    is.clear(std::ios::badbit | is.rdstate());
    std::cerr << "\nIntegerCong state description incomplete."
	      << "\nInput stream is probably mispositioned now." << std::endl;
  }
}
*/
