// $Id: Ranlux64Engine.h,v 1.1.1.1 2000/04/03 21:03:46 wilko Exp $
// -*- C++ -*-
//
// -----------------------------------------------------------------------
//                             HEP Random
//                        --- Ranlux64Engine ---
//                          class header file
// -----------------------------------------------------------------------
// The algorithm for this random engine has been taken from the notes of 
// a double-precision ranlux implementation by Martin Luscher, dated 
// November 1997.
//
// Like the previous ranlux generator, this one also has "luxury" levels,
// determining how many pseudo-random numbers are discarded for every 
// twelve values used. Three levels are given, with the note that Luscher
// himself advocates only the highest two levels for this engine.
//  level 0  (p=109):              Throw away 109 values for every 12 used
//  level 1  (p=202):   (default)  Throw away 202 values for every 12 used
//  level 2  (p=397):              Throw away 397 values for every 12 used
//
// The initialization is carried out using a Multiplicative Congruential
// generator using formula constants of L'Ecuyer as described in "F.James,
// Comp. Phys. Comm. 60 (1990) 329-344".
// =======================================================================
// Ken Smith      - Created Initial draft: 14th Jul 1998
//                - Added conversion operators:  6th Aug 1998
// =======================================================================

#ifndef Ranlux64Engine_h
#define Ranlux64Engine_h

#include "CLHEP/Random/RandomEngine.h"

class Ranlux64Engine : public HepRandomEngine {

public:

  Ranlux64Engine( std::istream& is );
  Ranlux64Engine();
  Ranlux64Engine( long seed, HepInt lux = 1 );
  Ranlux64Engine( HepInt rowIndex, HepInt colIndex, HepInt lux );
  virtual ~Ranlux64Engine();
  // Constructors and destructor

  Ranlux64Engine(const Ranlux64Engine &p);
  // Copy constructor

  Ranlux64Engine & operator = (const Ranlux64Engine &p);
  // Overloaded assignment operator, to retrieve the engine status.

  HepDouble flat();
  // It returns a pseudo random number between 0 and 1,
  // excluding the end points.

  void flatArray (const HepInt size, HepDouble* vect);
  // Fills the array "vect" of specified size with flat random values.

  void setSeed(long seed, HepInt lux=1);
  // Sets the state of the algorithm according to seed.

  void setSeeds(const long * seeds, HepInt lux=1);
  // Sets the state of the algorithm according to the zero terminated
  // array of seeds. Only the first seed is used.

  void saveStatus( const char filename[] = "Ranlux64.conf" ) const;
  // Saves in named file the current engine status.

  void restoreStatus( const char filename[] = "Ranlux64.conf" );
  // Reads from named file the last saved engine status and restores it.

  void showStatus() const;
  // Dumps the engine status on the screen.

  HepInt getLuxury() const { return luxury; }
  // Gets the luxury level.

  friend std::ostream& operator<< (std::ostream& os, const Ranlux64Engine& e);
  friend std::istream& operator>> (std::istream& is,       Ranlux64Engine& e);

private:

  HepInt nskip, luxury;
  HepInt i_lag, j_lag;
  HepInt count12;
  HepDouble seed_table[12];
  HepDouble carry;

  const HepDouble mantissa_bit_31;
  const HepDouble mantissa_bit_53;
  const HepDouble mantissa_bit_54;
  const HepDouble mantissa_bit_106;
  static HepInt numEngines;	
  static HepInt maxIndex;

}; // Ranlux64Engine

#endif // Ranlux64Engine_h
