// $Id: RanshiEngine.cpp,v 1.1.1.1 2000/04/03 21:03:46 wilko Exp $
// -*- C++ -*-
//
// -----------------------------------------------------------------------
//                           HEP Random
//                      --- RanshiEngine ---
//                    class implementation file
// -----------------------------------------------------------------------
//
// This algorithm implements the random number generator as proposed by 
// "F. Gutbrod, Comp. Phys. Comm. 87 (1995) 291-306". 
//
// =======================================================================
// Ken Smith      - Created:                                 9th June 1998
//                - Removed pow() from flat method:          21st Jul 1998
//                - Added conversion operators:               6th Aug 1998
// J. Marraffino  - Added some explicit casts to deal with
//                  machines where sizeof(int) != sizeof(long) 22 Aug 1998
// =======================================================================

#include "CLHEP/Random/RanshiEngine.h"
//#include "CLHEP/String/Strings.h"
#include <iomanip>

// Number of instances with automatic seed selection
HepInt RanshiEngine::numEngines = 0;

RanshiEngine::RanshiEngine() 
: mantissa_bit_32 ( pow(0.5,32. ) ),
  mantissa_bit_53 ( pow(0.5,53. ) ),
  mantissa_bit_54 ( pow(0.5,54. ) ),
  mantissa_bit_106( pow(0.5,106.) ),
  halfBuff(0), numFlats(0) 
{
  HepInt i = 0;
  long bvalue;
  while (i < numBuff) {    
    bvalue = numEngines+19780503L*(++i);
    buffer[i] = (unsigned int)bvalue;
  }
  theSeed = numEngines+19780503L*++i;
  redSpin = (unsigned int)(theSeed & 0xffffffff);
  ++numEngines;
  for( i = 0; i < 100; ++i) flat();    // Warm-up by running thorugh 100 nums
}

RanshiEngine::RanshiEngine(std::istream& is)
: mantissa_bit_32 ( pow(0.5,32. ) ),
  mantissa_bit_53 ( pow(0.5,53. ) ),
  mantissa_bit_54 ( pow(0.5,54. ) ),
  mantissa_bit_106( pow(0.5,106.) ),
  halfBuff(0), numFlats(0) 
{
  is >> *this;
}

RanshiEngine::RanshiEngine(long seed)
: mantissa_bit_32 ( pow(0.5,32. ) ),
  mantissa_bit_53 ( pow(0.5,53. ) ),
  mantissa_bit_54 ( pow(0.5,54. ) ),
  mantissa_bit_106( pow(0.5,106.) ),
  halfBuff(0), numFlats(0) 
{
  for (HepInt i = 0; i < numBuff; ++i) {
    buffer[i] = (unsigned int)seed;
  }
  theSeed = seed;
  redSpin = (unsigned int)(theSeed & 0xffffffff);
  for (HepInt j = 0; j < numBuff*20; ++j) { // "warm-up" for engine to hit
    flat();                               //  every ball on average 20X.
  }
}

RanshiEngine::RanshiEngine(HepInt rowIndex, HepInt colIndex)
: mantissa_bit_32 ( pow(0.5,32. ) ),
  mantissa_bit_53 ( pow(0.5,53. ) ),
  mantissa_bit_54 ( pow(0.5,54. ) ),
  mantissa_bit_106( pow(0.5,106.) ),
  halfBuff(0), numFlats(0) 
{
  HepInt i = 0;
  long bvalue;
  while( i < numBuff ) {
    bvalue = (rowIndex + (++i)*(colIndex+8));
    buffer[i] = (unsigned int)bvalue;
  }
  theSeed = rowIndex;
  redSpin = colIndex & 0xffffffff;
  for( i = 0; i < 100; ++i) flat();    // Warm-up by running thorugh 100 nums
}

RanshiEngine::~RanshiEngine() { }

RanshiEngine::RanshiEngine( const RanshiEngine & p ) 
: mantissa_bit_32 ( pow(0.5,32. ) ),
  mantissa_bit_53 ( pow(0.5,53. ) ),
  mantissa_bit_54 ( pow(0.5,54. ) ),
  mantissa_bit_106( pow(0.5,106.) ),
  halfBuff(0), numFlats(0) 
{
  *this = p;
}

RanshiEngine & RanshiEngine::operator=( const RanshiEngine & p ) {
  if( this != &p ) {
    halfBuff = p.halfBuff;
    numFlats = p.numFlats;
    redSpin  = p.redSpin;
    for( HepInt i =0; i < numBuff; ++i) {
      buffer[i] = p.buffer[i];
    }
  }
  return *this;
}

HepDouble RanshiEngine::flat() {
  unsigned int redAngle = (((numBuff/2) - 1) & redSpin) + halfBuff;
  unsigned int blkSpin     = buffer[redAngle] & 0xffffffff;
  unsigned int boostResult = blkSpin ^ redSpin;

  buffer[redAngle] = ((blkSpin << 17) | (blkSpin >> (32-17))) ^ redSpin;
  
  redSpin  = (blkSpin + numFlats++) & 0xffffffff;
  halfBuff = numBuff/2 - halfBuff;
  
  return ( blkSpin * mantissa_bit_32 +            // most significant part
	   (boostResult>>11) * mantissa_bit_53 +  // fill in remaining bits
	   mantissa_bit_54 - mantissa_bit_106 );  // non-zero
}

void RanshiEngine::flatArray(const HepInt size, HepDouble* vect) {
  for (HepInt i = 0; i < size; ++i) {
    vect[i] = flat();
  }
}

void RanshiEngine::setSeed(long seed, HepInt) {
  *this = RanshiEngine(seed); 
}

void RanshiEngine::setSeeds(const long* seeds, HepInt) {
  if (*seeds) {
    HepInt i = 0;
    while (seeds[i] && i < numBuff) {
      buffer[i] = (unsigned int)seeds[i];
      ++i;
    }
    while (i < numBuff) {
      buffer[i] = buffer[i-1];
      ++i;
    }
    theSeed = seeds[0];
    redSpin = (unsigned int)theSeed;
  }
  theSeeds = seeds;
}
     
void RanshiEngine::saveStatus(const char filename[]) const {
  std::ofstream outFile(filename, std::ios::out);
  if (!outFile.bad()) {
    outFile << std::setprecision(20) << theSeed << std::endl;
    for (HepInt i = 0; i < numBuff; ++i) {
      outFile << buffer[i] << " ";
    }
    outFile << redSpin  << " " << numFlats << " " << halfBuff << std::endl;
  }
}

void RanshiEngine::restoreStatus(const char filename[]) {
  std::ifstream inFile(filename, std::ios::in);
  if (!inFile.bad()) {
    inFile >> theSeed;
    for (HepInt i = 0; i < numBuff; ++i) {
      inFile >> buffer[i];
    }
    inFile >> redSpin >> numFlats >> halfBuff;
  }
}

void RanshiEngine::showStatus() const {
  std::cout << std::setprecision(20) << std::endl;
  std::cout << "----------- Ranshi engine status ----------" << std::endl;
  std::cout << "Initial seed      = " << theSeed << std::endl;
  std::cout << "Current red spin  = " << redSpin << std::endl;
  std::cout << "Values produced   = " << numFlats << std::endl;
  std::cout << "Side of buffer    = " << (halfBuff ? "upper" : "lower")
	    << std::endl;
  std::cout << "Current buffer    = " << std::endl;
  for (HepInt i = 0; i < numBuff; i+=4) {
    std::cout << std::setw(10) << std::setiosflags(std::ios::right)
	      << buffer[i]     << std::setw(11) << buffer[i+1] << std::setw(11)
	      << buffer[i+2]   << std::setw(11) << buffer[i+3] << std::endl;
  }
  std::cout << "-------------------------------------------" << std::endl;
}

RanshiEngine::operator HepFloat() {
  unsigned int redAngle = (((numBuff/2) - 1) & redSpin) + halfBuff;
  unsigned int blkSpin  = buffer[redAngle] & 0xffffffff;
  
  buffer[redAngle] = ((blkSpin << 17) | (blkSpin >> (32-17))) ^ redSpin;
  
  redSpin  = (blkSpin + numFlats++) & 0xffffffff;
  halfBuff = numBuff/2 - halfBuff;
  
  return HepFloat(blkSpin * mantissa_bit_32);
}

RanshiEngine::operator unsigned int() {
  unsigned int redAngle = (((numBuff/2) - 1) & redSpin) + halfBuff;
  unsigned int blkSpin  = buffer[redAngle] & 0xffffffff;
  
  buffer[redAngle] = ((blkSpin << 17) | (blkSpin >> (32-17))) ^ redSpin;
  
  redSpin  = (blkSpin + numFlats++) & 0xffffffff;
  halfBuff = numBuff/2 - halfBuff;
  
  return blkSpin;
}
/*
std::ostream& operator<< (std::ostream& os, const RanshiEngine& e) {
  HepString beginMarker = "RanshiEngine-begin";
  HepString endMarker   = "RanshiEngine-end";
 
  os << " " << beginMarker << " ";
  
  os << std::setprecision(20)  << e.theSeed  << " ";
  for (HepInt i = 0; i < e.numBuff; ++i) {
    os << e.buffer[i]  << " ";
  }
  os << e.redSpin  << " " << e.numFlats << " " << e.halfBuff; 
  
  os << " " << endMarker   << " ";
  return os;
}
*/
/*
std::istream& operator>> (std::istream& is, RanshiEngine& e) {
  HepString beginMarker = " ";
  HepString endMarker   = " ";
  
  is >> std::ws >> beginMarker;
  if (beginMarker != "RanshiEngine-begin") {
    is.clear(std::ios::badbit | is.rdstate());
    std::cerr << "\nInput mispositioned or"
	      << "\nRanshiEngine state description missing or"
	      << "\nwrong engine type found." << std::endl;
    return is;
  }
  is >> e.theSeed;
  for (HepInt i = 0; i < e.numBuff; ++i) {
    is >> e.buffer[i];
  }
  is >> e.redSpin >> e.numFlats >> e.halfBuff;
  is >> std::ws >> endMarker;
  if (endMarker != "RanshiEngine-end") {
    is.clear(std::ios::badbit | is.rdstate());
    std::cerr << "\nRanshiEngine state description incomplete."
	      << "\nInput stream is probably mispositioned now." << std::endl;
    return is;
  }
  return is;
}
*/
