// $Id: TripleRand.cpp,v 1.1.1.1 2000/04/03 21:03:46 wilko Exp $
// -*- C++ -*-
//
// -----------------------------------------------------------------------
//                           Hep Random
//                       --- TripleRand ---
//                   class implementation file
// -----------------------------------------------------------------------
// A canopy pseudo-random number generator. Using the Tausworthe
// exclusive-or shift register, a simple Integer Coungruence generator, and
// the Hurd 288 total bit shift register, all XOR'd with each other, we
// provide an engine that should be a fairly good "mother" generator.
// =======================================================================
// Ken Smith      - Initial draft started:                    23rd Jul 1998
//                - Added conversion operators:                6th Aug 1998
// J. Marraffino  - Added some explicit casts to deal with
//                  machines where sizeof(int) != sizeof(long)  22 Aug 1998
// =======================================================================

#include "CLHEP/Random/TripleRand.h"
//#include "CLHEP/String/Strings.h"
#include <iomanip>

// Number of instances with automatic seed selection
HepInt TripleRand::numEngines = 0;

//********************************************************************
//   TripleRand  
//********************************************************************

TripleRand::TripleRand() 
: mantissa_bit_32 ( pow(0.5,32. ) ),
  mantissa_bit_53 ( pow(0.5,53. ) ),
  mantissa_bit_54 ( pow(0.5,54. ) ),
  mantissa_bit_106( pow(0.5,106.) ),
  tausworthe (1234567 + numEngines + 175321),
  integerCong(69607 * tausworthe + 54329, numEngines),
  hurd(19781127 + integerCong)
{
  theSeed = 1234567;
  ++numEngines;
}

TripleRand::TripleRand(long seed)
: mantissa_bit_32 ( pow(0.5,32. ) ),
  mantissa_bit_53 ( pow(0.5,53. ) ),
  mantissa_bit_54 ( pow(0.5,54. ) ),
  mantissa_bit_106( pow(0.5,106.) ),
  tausworthe ((unsigned int)seed + numEngines + 175321),
  integerCong(69607 * tausworthe + 54329, numEngines),
  hurd(19781127 + integerCong)
{
  theSeed = seed;
}

TripleRand::TripleRand(std::istream & is) 
: mantissa_bit_32 ( pow(0.5,32. ) ),
  mantissa_bit_53 ( pow(0.5,53. ) ),
  mantissa_bit_54 ( pow(0.5,54. ) ),
  mantissa_bit_106( pow(0.5,106.) )
{
  is >> *this;
}

TripleRand::TripleRand(HepInt rowIndex, HepInt colIndex)
: mantissa_bit_32 ( pow(0.5,32. ) ),
  mantissa_bit_53 ( pow(0.5,53. ) ),
  mantissa_bit_54 ( pow(0.5,54. ) ),
  mantissa_bit_106( pow(0.5,106.) ),
  tausworthe (rowIndex + numEngines * colIndex + 175321),
  integerCong(69607 * tausworthe + 54329, numEngines),
  hurd(19781127 + integerCong)
{
  theSeed = rowIndex;
}

TripleRand::~TripleRand() { }

TripleRand::TripleRand( const TripleRand & p )
: mantissa_bit_32 ( pow(0.5,32. ) ),
  mantissa_bit_53 ( pow(0.5,53. ) ),
  mantissa_bit_54 ( pow(0.5,54. ) ),
  mantissa_bit_106( pow(0.5,106.) )
{
  *this = p;
}

TripleRand & TripleRand::operator=( const TripleRand & p ) {
  if( this != &p ) {
     tausworthe  = p.tausworthe;
     integerCong = p.integerCong;
     hurd        = p.hurd;
  }
  return *this;
}

HepDouble TripleRand::flat() {
  unsigned int ic ( integerCong );
  unsigned int t  ( tausworthe  );
  unsigned int h  ( hurd        );
  return ( (t  ^ ic ^ h) * mantissa_bit_32 +          // most significant part
               (t >> 11) * mantissa_bit_53 +          // fill in remaining bits
           mantissa_bit_54 - mantissa_bit_106         // make sure non-zero
         );
}

void TripleRand::flatArray(const HepInt size, HepDouble* vect) {
  for (HepInt i = 0; i < size; ++i) {
    vect[i] = flat();
  }
}

void TripleRand::setSeed(long seed, HepInt) {
  theSeed = seed;
  tausworthe  = Tausworthe((unsigned int)seed + numEngines + 175321);
  integerCong = IntegerCong(69607 * tausworthe + 54329, numEngines);
  hurd        = Hurd288Engine( 19781127 + integerCong );
}

void TripleRand::setSeeds(const long * seeds, HepInt) {
  setSeed(seeds ? *seeds : 1234567, 0);
  theSeeds = seeds;
}

void TripleRand::saveStatus(const char filename[]) const {
  std::ofstream outFile(filename, std::ios::out);
  if (!outFile.bad()) {
    outFile << std::setprecision(20) << theSeed << " ";
    outFile << ConstTaus() << " " 
            << ConstCong() << " " 
            << ConstHurd() << " ";
  }
}

void TripleRand::restoreStatus(const char filename[]) {
  std::ifstream inFile(filename, std::ios::in);
  if (!inFile.bad()) {
    inFile >> theSeed >> Taus() >> Cong() >> Hurd();
  }
}

void TripleRand::showStatus() const {
  std::cout << std::setprecision(20) << std::endl;
  std::cout << "-------- TripleRand engine status ---------" << std::endl;
  std::cout << "Initial seed           = " << theSeed << std::endl;
  std::cout << "Tausworthe generator   = " << std::endl << ConstTaus();
  std::cout << "IntegerCong generator  = " << std::endl << ConstCong();
  std::cout << "Hurd288Engine generator= " << std::endl << ConstHurd();
  std::cout << std::endl
	    << "-----------------------------------------" << std::endl;
}

TripleRand::operator float() {
  return (HepFloat)(( integerCong ^ tausworthe ^ hurd ) * mantissa_bit_32 );
}

TripleRand::operator unsigned int() {
  return integerCong ^ tausworthe ^ hurd;
}

Tausworthe    & TripleRand::Taus() { return tausworthe;  }
IntegerCong   & TripleRand::Cong() { return integerCong; }
Hurd288Engine & TripleRand::Hurd() { return hurd;        }

const Tausworthe    & TripleRand::ConstTaus() const { return tausworthe;  }
const IntegerCong   & TripleRand::ConstCong() const { return integerCong; }
const Hurd288Engine & TripleRand::ConstHurd() const { return hurd;        }

/*
std::ostream & operator<< (std::ostream & os, const TripleRand & e) {
  HepString beginMarker = "TripleRand-begin";
  HepString endMarker   = "TripleRand-end";

  os << " " << beginMarker << " ";
  os << std::setprecision(20) << e.theSeed << " "
     << e.ConstTaus() << " " << e.ConstCong() << " " << e.ConstHurd();
  os << " " <<  endMarker  << " ";
  return os;
}
*/
/*
std::istream & operator>> (std::istream & is, TripleRand & e) {
  HepString beginMarker = " ";
  HepString endMarker   = " ";

  is >> std::ws >> beginMarker;
  if (beginMarker != "TripleRand-begin") {
    is.clear(std::ios::badbit | is.rdstate());
    std::cerr << "\nInput mispositioned or"
	      << "\nTripleRand state description missing or"
	      << "\nwrong engine type found." << std::endl;
    return is;
  }
  is >> e.theSeed >> e.Taus() >> e.Cong() >> e.Hurd();
  is >> std::ws >> endMarker;
  if (endMarker != "TripleRand-end") {
    is.clear(std::ios::badbit | is.rdstate());
    std::cerr << "\nTripleRand state description incomplete."
	      << "\nInput stream is probably mispositioned now." << std::endl;
    return is;
  }
  return is;
}
*/

//********************************************************************
//   Tausworthe
//********************************************************************

Tausworthe::Tausworthe() {
  words[0] = 1234567;
  for (wordIndex = 1; wordIndex < 4; ++wordIndex) {
    words[wordIndex] = 69607 * words[wordIndex-1] + 54329;
  } 
}

Tausworthe::Tausworthe(unsigned int seed) {
  words[0] = seed;
  for (wordIndex = 1; wordIndex < 4; ++wordIndex) {
    words[wordIndex] = 69607 * words[wordIndex-1] + 54329;
  }
}

Tausworthe::operator unsigned int() {

// Mathematically:  Consider a sequence of bits b[n].  Repeatedly form
// b[0]' = b[127] ^ b[97]; b[n]' = b[n-1].  This sequence will have a very
// long period (2**127-1 according to Tausworthe's work).

// The actual method used relies on the fact that the operations needed to
// form bit 0 for up to 96 iterations never depend on the results of the
// previous ones.  So you can actually compute many bits at once.  In fact
// you can compute 32 at once -- despite 127 - 97 < 32 -- but 24 was used in
// the method used in Canopy, where they wanted only single-precision float
// randoms.  I will do 32 here.

// When you do it this way, this looks disturbingly like the dread lagged XOR
// Fibonacci.  And indeed, it is a lagged Fibonacii, F(4,3, op) with the op
// being the XOR of a combination of shifts of the two numbers.  Although
// Tausworthe asserted excellent properties, I would be scared to death.
// However, the shifting and bit swapping really does randomize this in a
// serious way.

// Statements have been made to the effect that shift register sequences fail
// the parking lot test because they achieve randomness by multiple foldings,
// and this produces a characteristic pattern.  We observe that in this
// specific algorithm, which has a fairly long lever arm, the foldings become
// effectively random.  This is evidenced by the fact that the generator
// does pass the Diehard tests, including the parking lot test.

// To avoid shuffling of variables in memory, you either have to use circular
// pointers (and those give you ifs, which are also costly) or compute at least
// a few iterations at once.  We do the latter.  Although there is a possible
// trade of room for more speed, by computing and saving 256 instead of 128
// bits at once, I will stop at this level of optimization.

// To remind:  Each (32-bit) step takes the XOR of bits [127-96] with bits
// [95-64] and places it in bits [0-31].  But in the first step, we designate
// word0 as bits [0-31], in the second step, word 1 (since the bits it holds
// will no longer be needed), then word 2, then word 3.  After this, the
// stream contains 128 random bits which we will use as 4 valid 32-bit
// random numbers.

// Thus at the start of the first step, word[0] contains the newest (used)
// 32-bit random, and word[3] the oldest.   After four steps, word[0] again
// contains the newest (now unused) random, and word[3] the oldest.
// Bit 0 of word[0] is logically the newest bit, and bit 31 of word[3]
// the oldest.

  if (wordIndex <= 0) {
    for (wordIndex = 0; wordIndex < 4; ++wordIndex) {
      words[wordIndex] = ( (words[(wordIndex+1) & 3] << 1 ) |
                                   (words[wordIndex] >> 31)   )
                       ^ ( (words[(wordIndex+1) & 3] << 31) |
                                   (words[wordIndex] >>  1)   );
    }
  }
  return words[--wordIndex] & 0xffffffff;
}
/*
std::ostream & operator<< ( std::ostream & os, const Tausworthe & e ) {
  HepString beginMarker = "Tausworthe-begin";
  HepString endMarker   = "Tausworthe-end";

  os << " " << beginMarker << " ";
  os << std::setprecision(20);
  for (HepInt i = 0; i < 4; ++i) {
    os << e.words[i] << " ";
  }
  os << e.wordIndex;
  os << " " <<  endMarker  << " ";
  return os;
}
*/
/*
std::istream & operator>> ( std::istream & is, Tausworthe & e ) {
  HepString beginMarker = " ";
  HepString endMarker   = " ";

  is >> std::ws >> beginMarker;
  if (beginMarker != "Tausworthe-begin") {
    is.clear(std::ios::badbit | is.rdstate());
    std::cerr << "\nInput mispositioned or"
	      << "\nTausworthe state description missing or"
	      << "\nwrong engine type found." << std::endl;
    return is;
  }
  for (HepInt i = 0; i < 4; ++i) {
    is >> e.words[i];
  }
  is >> e.wordIndex;
  is >> std::ws >> endMarker;
  if (endMarker != "Tausworthe-end") {
    is.clear(std::ios::badbit | is.rdstate());
    std::cerr << "\nTausworthe state description incomplete."
	      << "\nInput stream is probably mispositioned now." << std::endl;
    return is;
  }
  return is;
}
*/
//********************************************************************
//   IntegerCong
//********************************************************************

IntegerCong::IntegerCong() 
: state((unsigned int)3758656018),
  multiplier(66565),
  addend(12341) 
{
}

IntegerCong::IntegerCong(unsigned int seed, HepInt streamNumber)
: state(seed),
  multiplier(65536 + 1024 + 5 + (8 * 1017 * streamNumber)),
  addend(12341)
{
  // As to the multiplier, the following comment was made:
  // We want our multipliers larger than 2^16, and equal to
  // 1 mod 4 (for max. period), but not equal to 1 mod 8 
  // (for max. potency -- the smaller and higher dimension the 
  // stripes, the better)

  // All of these will have fairly long periods.  Depending on the choice
  // of stream number, some of these will be quite decent when considered
  // as independent random engines, while others will be poor.  Thus these
  // should not be used as stand-alone engines; but when combined with a
  // generator known to be good, they allow creation of millions of good
  // independent streams, without fear of two streams accidentally hitting
  // nearby places in the good random sequence.
}

IntegerCong::operator unsigned int() {
  return state = (state * multiplier + addend) & 0xffffffff;
}
/*
std::ostream & operator<< ( std::ostream & os, const IntegerCong & e ) {
  HepString beginMarker = "IntegerCong-begin";
  HepString endMarker   = "IntegerCong-end";

  os << " " << beginMarker << " ";
  os << std::setprecision(20);
  os << e.state << " " << e.multiplier << " " << e.addend;
  os << " " <<  endMarker  << " ";
  return os;
}
*/
/*
std::istream & operator>> ( std::istream & is, IntegerCong & e ) {
  HepString beginMarker = " ";
  HepString endMarker   = " ";

  is >> std::ws >> beginMarker;
  if (beginMarker != "IntegerCong-begin") {
    is.clear(std::ios::badbit | is.rdstate());
    std::cerr << "\nInput mispositioned or"
	      << "\nIntegerCong state description missing or"
	      << "\nwrong engine type found." << std::endl;
    return is;
  }
  is >> e.state >> e.multiplier >> e.addend;
  is >> std::ws >> endMarker;
  if (endMarker != "IntegerCong-end") {
    is.clear(std::ios::badbit | is.rdstate());
    std::cerr << "\nIntegerCong state description incomplete."
	      << "\nInput stream is probably mispositioned now." << std::endl;
    return is;
  }
  return is;
}
*/
