// -*- C++ -*-
// $Id: Rotation.cpp,v 1.1.1.1 2000/04/03 21:03:47 wilko Exp $
// ---------------------------------------------------------------------------
//
// This file is a part of the CLHEP - a Class Library for High Energy Physics.
//
// This is the implementation of the HepRotation class.
//

#ifdef __GNUG__
#pragma implementation
#endif

#include "CLHEP/config/TemplateFunctions.h"
#include "CLHEP/Vector/Rotation.h"

#ifdef HEP_HAVE_NEW_IOSTREAMS
#include <iostream>
#else
#include <iostream.h>
#endif

#ifdef HEP_DEBUG_INLINE
#include "CLHEP/Vector/Rotation.inl"
#endif

HepDouble HepRotation::operator() (int i, int j) const {
  if (i == 0) {
    if (j == 0) { return xx(); }
    if (j == 1) { return xy(); }
    if (j == 2) { return xz(); } 
  } else if (i == 1) {
    if (j == 0) { return yx(); }
    if (j == 1) { return yy(); }
    if (j == 2) { return yz(); } 
  } else if (i == 2) {
    if (j == 0) { return zx(); }
    if (j == 1) { return zy(); }
    if (j == 2) { return zz(); } 
  } 
  std::cerr << "HepRotation subscripting: bad indeces "
       << "(" << i << "," << j << ")" << std::endl;
  return 0.0;
} 

HepRotation HepRotation::operator* (const HepRotation & b) const {
  return HepRotation(xx()*b.xx() + xy()*b.yx() + xz()*b.zx(),
		     xx()*b.xy() + xy()*b.yy() + xz()*b.zy(),
		     xx()*b.xz() + xy()*b.yz() + xz()*b.zz(),
		     yx()*b.xx() + yy()*b.yx() + yz()*b.zx(),
		     yx()*b.xy() + yy()*b.yy() + yz()*b.zy(),
		     yx()*b.xz() + yy()*b.yz() + yz()*b.zz(),
		     zx()*b.xx() + zy()*b.yx() + zz()*b.zx(),
		     zx()*b.xy() + zy()*b.yy() + zz()*b.zy(),
		     zx()*b.xz() + zy()*b.yz() + zz()*b.zz());
}

HepRotation & HepRotation::rotate(HepDouble a, const Hep3Vector& axis) {
  if (a != 0.0) {
    HepDouble ll = axis.mag();
    if (ll == 0.0) {
      //      std::cerr << "HepRotation: zero axis" << std::endl;
    }else{
      HepDouble sa = sin(a), ca = cos(a);
      HepDouble dx = axis.x()/ll, dy = axis.y()/ll, dz = axis.z()/ll;   
      HepRotation m(
	ca+(1-ca)*dx*dx,          (1-ca)*dx*dy-sa*dz,    (1-ca)*dx*dz+sa*dy,
	   (1-ca)*dy*dx+sa*dz, ca+(1-ca)*dy*dy,          (1-ca)*dy*dz-sa*dx,
	   (1-ca)*dz*dx-sa*dy,    (1-ca)*dz*dy+sa*dx, ca+(1-ca)*dz*dz );
      transform(m);
    }
  }
  return *this;
}

HepRotation & HepRotation::rotateX(HepDouble a) {
  HepDouble c = cos(a);
  HepDouble s = sin(a);
  HepDouble x = ryx, y = ryy, z = ryz; 
  ryx = c*x - s*rzx;
  ryy = c*y - s*rzy;
  ryz = c*z - s*rzz;
  rzx = s*x + c*rzx;
  rzy = s*y + c*rzy;
  rzz = s*z + c*rzz;
  return *this;
}

HepRotation & HepRotation::rotateY(HepDouble a){
  HepDouble c = cos(a);
  HepDouble s = sin(a);
  HepDouble x = rzx, y = rzy, z = rzz; 
  rzx = c*x - s*rxx;
  rzy = c*y - s*rxy;
  rzz = c*z - s*rxz;
  rxx = s*x + c*rxx;
  rxy = s*y + c*rxy;
  rxz = s*z + c*rxz;
  return *this;
}

HepRotation & HepRotation::rotateZ(HepDouble a) {
  HepDouble c = cos(a);
  HepDouble s = sin(a);
  HepDouble x = rxx, y = rxy, z = rxz; 
  rxx = c*x - s*ryx;
  rxy = c*y - s*ryy;
  rxz = c*z - s*ryz;
  ryx = s*x + c*ryx;
  ryy = s*y + c*ryy;
  ryz = s*z + c*ryz;
  return *this;
}

HepRotation & HepRotation::rotateAxes(const Hep3Vector &newX,
				      const Hep3Vector &newY,
				      const Hep3Vector &newZ) {
  HepDouble del = 0.001;
  Hep3Vector w = newX.cross(newY);

  if (abs(newZ.x()-w.x()) > del ||
      abs(newZ.y()-w.y()) > del ||
      abs(newZ.z()-w.z()) > del ||
      abs(newX.mag2()-1.) > del ||
      abs(newY.mag2()-1.) > del || 
      abs(newZ.mag2()-1.) > del ||
      abs(newX.dot(newY)) > del ||
      abs(newY.dot(newZ)) > del ||
      abs(newZ.dot(newX)) > del) {
    std::cerr << "HepRotation::rotateAxes: bad axis vectors" << std::endl;
    return *this;
  }else{
    return transform(HepRotation(newX.x(), newY.x(), newZ.x(),
                                 newX.y(), newY.y(), newZ.y(),
                                 newX.z(), newY.z(), newZ.z()));
  }
}

HepDouble HepRotation::phiX() const {
  return (yx() == 0.0 && xx() == 0.0) ? 0.0 : atan2(yx(),xx());
}

HepDouble HepRotation::phiY() const {
  return (yy() == 0.0 && xy() == 0.0) ? 0.0 : atan2(yy(),xy());
}

HepDouble HepRotation::phiZ() const {
  return (yz() == 0.0 && xz() == 0.0) ? 0.0 : atan2(yz(),xz());
}

HepDouble HepRotation::thetaX() const {
  return acos(zx());
}

HepDouble HepRotation::thetaY() const {
  return acos(zy());
}

HepDouble HepRotation::thetaZ() const {
  return acos(zz());
}

void HepRotation::getAngleAxis(HepDouble &angle, Hep3Vector &axis) const {
  HepDouble cosa  = 0.5*(xx()+yy()+zz()-1);
  HepDouble cosa1 = 1-cosa;
  if (cosa1 <= 0) {
    angle = 0;
    axis  = Hep3Vector(0,0,1);
  }else{
    HepDouble x=0, y=0, z=0;
    if (xx() > cosa) x = sqrt((xx()-cosa)/cosa1);
    if (yy() > cosa) y = sqrt((yy()-cosa)/cosa1);
    if (zz() > cosa) z = sqrt((zz()-cosa)/cosa1);
    if (zy() < yz())  x = -x;
    if (xz() < zx())  y = -y;
    if (yx() < xy())  z = -z;
    angle = acos(cosa);
    axis  = Hep3Vector(x,y,z);
  }
}
