// -*- C++ -*-
// $Id: ThreeVector.cpp,v 1.1.1.1 2000/04/03 21:03:47 wilko Exp $
// ---------------------------------------------------------------------------
//
// This file is a part of the CLHEP - a Class Library for High Energy Physics.
//
// This is the implementation of the Hep3Vector class.
//

#ifdef __GNUG__
#pragma implementation
#endif

#include "CLHEP/Vector/ThreeVector.h"
#include "CLHEP/Vector/Rotation.h"

//#include "CLHEP/String/Strings.h"
#include <CLHEP/config/strstream.h>
#include <ctype.h>

#ifdef HEP_DEBUG_INLINE
#include "CLHEP/Vector/ThreeVector.inl"
#endif

HepDouble Hep3Vector::operator () (int i) const {
  switch(i) {
  case X:
    return x();
  case Y:
    return y();
  case Z:
    return z();
  default:
    std::cerr << "Hep3Vector subscripting: bad index (" << i << ")"
	      << std::endl;
  }
  return 0.;
}

HepDouble & Hep3Vector::operator () (int i) {
  switch(i) {
  case X:
    return dx;
  case Y:
    return dy;
  case Z:
    return dz;
  default:
    std::cerr << "Hep3Vector subscripting: bad index (" << i << ")"
	      << std::endl;
  }
  HepDouble dummy;
  HepDouble & rdummy = dummy;
  return rdummy;
}

Hep3Vector & Hep3Vector::operator *= (const HepRotation & m){
  return *this = m * (*this);
}

Hep3Vector & Hep3Vector::transform(const HepRotation & m) {
  return *this = m * (*this);
}

void Hep3Vector::rotateX(HepDouble angle) {
  HepDouble s = sin(angle);
  HepDouble c = cos(angle);
  HepDouble yy = dy;
  dy = c*yy - s*dz;
  dz = s*yy + c*dz;
}

void Hep3Vector::rotateY(HepDouble angle) {
  HepDouble s = sin(angle);
  HepDouble c = cos(angle);
  HepDouble zz = dz;
  dz = c*zz - s*dx;
  dx = s*zz + c*dx;
}

void Hep3Vector::rotateZ(HepDouble angle) {
  HepDouble s = sin(angle);
  HepDouble c = cos(angle);
  HepDouble xx = dx;
  dx = c*xx - s*dy;
  dy = s*xx + c*dy;
}

void Hep3Vector::rotate(HepDouble angle, const Hep3Vector & axis){
  HepRotation trans;
  trans.rotate(angle, axis);
  operator*=(trans);
}

void Hep3Vector::rotateUz(const Hep3Vector& NewUzVector) {
  // NewUzVector must be normalized !

  HepDouble u1 = NewUzVector.x();
  HepDouble u2 = NewUzVector.y();
  HepDouble u3 = NewUzVector.z();
  HepDouble up = u1*u1 + u2*u2;

  if (up) {
      up = sqrt(up);
      HepDouble px = dx,  py = dy,  pz = dz;
      dx = (u1*u3*px - u2*py + u1*up*pz)/up;
      dy = (u2*u3*px + u1*py + u2*up*pz)/up;
      dz = (u3*u3*px -    px + u3*up*pz)/up;
    }
  else if (u3 < 0.) { dx = -dx; dz = -dz; }      // phi=0  teta=pi
  else {};
}

HepDouble Hep3Vector::pseudoRapidity() const {
  HepDouble m = mag();
  return 0.5*log( (m+z())/(m-z()) );
}

std::ostream & operator << (std::ostream & s, const Hep3Vector & q) {
  return s << "(" << q.x() << "," << q.y() << "," << q.z() << ")";
}

#if 0
std::istream & operator >> (std::istream & s, Hep3Vector & q) {
  // Parse:
  HepString cleanString;
  char c;
  int numRead = 0;
  HepBoolean previosIsSpace = false;

  while(EOF != (c = s.get()) && c != ')' && s.good()) {
    // ignore initial bracket or spaces
    if (!cleanString.isEmpty() || ('(' != c && 0 == isspace(c))){            
      if (',' == c){          // replace ',' with ' '
	c = ' ';
      }
      cleanString += c;     // take c
      if (isspace(c)){      // count # of white spaces (excluding initials)
	if (false == previosIsSpace) {
	  ++numRead;
	}
	previosIsSpace = true;
	if (numRead >= q.SIZE){  // read enough
	  break;
	}
      }
      else {
	previosIsSpace = false;
      }
    }
  }

  std::istrstream newStream(cleanString);
  double xx(q.x()), yy(q.y()), zz(q.z());
  newStream >> xx >> yy >> zz;
  q.setX(xx);
  q.setY(yy);
  q.setZ(zz);

  return s;
}
#endif

const Hep3Vector HepXHat(1.0, 0.0, 0.0);
const Hep3Vector HepYHat(0.0, 1.0, 0.0);
const Hep3Vector HepZHat(0.0, 0.0, 1.0);
