#ifndef selectionManager_H
#define selectionManager_H

#include "Event/cutVI.h"
#include "Event/cutSelectionTask.h"

#include "Event/serviceI.h"
#include "Event/serverVI.h"

//---------------------
//! stores, servers and applies the system and user defined cuts 
/*!
\e selectionManager stores the system and user defined cuts. <br>
\e selectionManager returns the cuts and cutComposite via the \c getCut() method 
	and \c getCutComposite() methods<br>
\e selectionManager applies a cut via its \c apply() method.<br>
General features:
<ul>
	<li> the <b> user defined cuts </b> should be created in the userCuts class.
	<li> the <b> user defined cuts </b> should be added to the server using the
	method \c addCut().
	<li> the method \c getCut() returns the a cut (cutVI) with name nameCut from
	the server of the selectionManager. These cuts are either simple cuts (cutVI) or multiple
	cuts (cutComposite). 
	<li> In addition, the method \c apply() returns the result of applying
	the cut.
	<li> the method \c getCutComposite() returns the cutComposite with name nameCuts.
	the cutComposite is a multicut that allows access
	to everyone of the cuts by name or index. (see cutComposite for details).<br>
</ul>
Centella features:
<ul>
	<li> selectionManager has a protected server with cutSelectionTask. 
	cutSelectionTasks are just  a cutComposite with serviceI services. <br> 
	cutSelectionTasks have a object name to refer to them.
	they have the possibility of adding cuts into the cutComposite,
	that implies that an user can dinamically define a cutSelectionTask
	by adding particular cuts.<br>
	A cutSelectionTask can be added using \c addCutSelectionTask() and retreived
	via \c getCutSelectionTask().
	<li> An external user (optionManager) can define a new cutSelectionTask and from there
	add cuts into the cut Selection.
	<li> The \e Centella implementation of selectionServer have two already created
	cutSelectionTask: they are the \e "selRead" and the \e "selWrite" selections. 
	<ul> 
		<li> The \e "selRead" selection contains the selection criteria or cuts required to read the event
		, from taking it from persistency.
		<li> The \e "selWrite" contains the selection criteria 
		that a processed event should fulfill in order to be written (saved) 
		into persistency.
	</ul>
</ul>
*/
//----------------------------------------
//  JA hernando   Santa Cruz, CA, 052600
//----------------------------------------

//########################################
class selectionManager : public serviceI
//########################################
{
public:
	
	
	//! default constructor - constructs the cutUserServer
	selectionManager();
	//! destructor 
	~selectionManager();

	//! Singleton: returns the Manager
	static selectionManager* instance() {return m_instance;}
	
	//! apply the named cutVI
	bool apply(std::string name)   {return getCut(name)->apply();}

	//! returns a named cutVI
	cutVI* getCut(std::string name) const;
	//! add a cut into the server
	void addCut(std::string name, cutVI* cut);

	//! returns a cutComposite
	cutComposite* getCutComposite(std::string name) {return getCutSelectionTask(name);}
	
	//! returns a named cutSelectionTask
	cutSelectionTask* getCutSelectionTask(std::string name) const;
	//! add a cutSelectionTask into the server
	void addCutSelectionTask(std::string name, cutSelectionTask* sel);

	//! info of the manager
	virtual void writeOut() const;

protected:

	//! define the options optionVI
	virtual void defineOption();

	//! the User define the selections (cutComposite) via optionVI
	virtual void setOption(std::string selectionName, std::string cutName); //overwriten
	
private:

	//! Singleton: pointer to the class
	static selectionManager* m_instance;

	//! Pointer to the user defined cut in the Server
	serverVI<cutVI>* m_serverCut;

	//! Pointer to the selections in the Server
	serverVI<cutSelectionTask>* m_serverSelection;

private:

	//! dummy string for the options
	std::string m_selName;
};
#endif
