
#ifndef __centella_H
#define __centella_H

//----------------------------------------------
//
//   CENTELLA
//
//----------------------------------------------
//             J.A Hernando, Santa Cruz 02/29/00
//----------------------------------------------

class processManager;
class dataManager;

class messageManager;
class optionManager;
class selectionManager;

#include "Event/algorithmTree.h"
#include "Event/defineVI.h"

//! Application class
/*! 
	\e Centella creates the \b system managers: dataManager, processManager, selectionManager, 
	optionManager and messageManager. <br>
	\e Centella executes the program via its algorithmTree base class. <br>
	\e Centella is statically implemented in the code of its \c define() method. <br>
	<ul>
	<li> The system has five managers. Three should have a \e Centella specific
	implementation:	dataManager, processManager, selectionManager.
	Which declares the \b data (persistent and transient) the \b processes (algorithms and
	multialgorithms) and \b selections (cuts and multicuts). These managers 
	declare and use the data, algorithms and selections relevant for the
	\e centella application. The virtual base classes
	for data, algorithms and selections are trsDataVI, algorithmVI and cutVI.
	<br><ul>
	<br><li>The data is clasified into detector data -geometry and calibration- (dataUserServerDet);
	event data - data relevant for every event - (dataUserServerEvent) and
	persistend data (dataUserServerROOT). In \e Centella the persisten data is
	stored into ROOT trees, histograms and ntuples. The dataManager serves the
	transient data via the \c getData() template method. In \e Centella calibration
	and event data are declared transient. The other data, persistent
	and geometry is for the moment provided via pointers. Most of the data is
	created one, it reference pointer stored into the server and its contents
	update. 
	<br><li> The processes are algorithms. They are some system process and some user defined ones.
	The algorithms can be simple (algorithmVI, algorithmCommand) or multialgorithms 
	(algorithmComposite, algorithmTree). Their main method is \c run(). The processManager serves the algorithms
	vita the \c getAlgorithm() method.
	In the processManager there are servers to algorithmTask and eventProcess which are
	algorithms that can be defined dinamically, in particular, they can be created
	by the user via the input file and the optionManager \c setOption() method. The algorithms
	are created at one and not removed after stored into the server. 
	<br><li> The selections are cuts. They are some system cuts and some defined by the user. 
	They are simple cuts (cutVI) or multicuts (cutComposite). Their main method is
	\c apply(). The selectionManager
	serves the selection via the \c getCut() method. selectionManager has also a server
	for cutSelection, a cutComposite that can be defined by the user. Again, the cuts
	should be construct an included in the server and not deleted.
	<br></ul><br>
	There are two independent managers: optionManager and messageManager. 
	<br><ul>
	<br><li> The optionManager serves to set options (that can be turn into operations) 
	into servived classes derived from the base class optionVI. Previously, paramaters of the classes,
	(types accepted: strings, double, int) should be turn into options 
	using the method \c defineOption() declared in optionVI. In this way it is possible to create
	objects dinamically, in particular algorithms and selections by the user. 
	In \e Centella algorithms and selections are dinamically implemented via 
	the input file. 
	<br><li>The messageManager defines an ostream where to send
	the messages and provides some methods (i.e \c message()) to send them. 
	The messageManager and
	the classes serviced by messageVI has a message level that determines if a message should
	or not realized.
	</ul><br>
	<br><li> Centella is an algorithmTree that runs: 1) the execution of some initial algorithms
	2) the execution of a run, which loop in a collection of events to apply some 
	algorithms, 3) the execution of some final algorithms.
	<br><ul>
	<br><li> Some of the algorithms executed in the initial and finalize part
	are already declared in \c define(), except two of them. They are the algorithmTask 
	named \e "iniOfRun" and \e "endOfRun" where the \b user can define the algorithms 
	to execute at the start and the end of the run. 
	<br><li> The execute() method of
	\e centella run the runRunAlg algorithm, a system algorithm that performs the loop on
	the number of event of the persistent storage and call the algorithm runEventAlg to
	process every event. 
	<br><li>The runEventAlg has an eventProcess (algorithm) called \e "runEvent" (a eventProcess) 
	which the user can indicate which algorithms to run dinamically to. In addition
	runEventAlg has two cutSelection named "selRead" and "selAna" which decide if
	the event shoulb be processed and if the processed event should be intalled in
	persistent data. This two cutSelection can be dinamically defined by the user.
	</ul>
	</ul>
*/
//-----------------------------------------------
//   J.A Hernando   Santa Cruz, CA, 06/06/00
//-----------------------------------------------
//###############################################
class centella : public algorithmTree , public defineVI
//###############################################
{

public:


	//! constructor - initalize managers
	centella();
	//! destructors - delete managers
	~centella();

	//! define the class
	virtual void define();

	//! centella is a Singleton (to be use by centellaGUI). Access to the pointer
	static centella* instance() {return m_instance;}

private:

	//! singleton: pointer to the class
	static centella*   m_instance;

	//! application independent messageManager
	messageManager*   m_messageManager;
	//! application independent optionManager
	optionManager*    m_optionManager;
	
	//! application dependent dataManager
	dataManager*        m_dataManager;
	//! application dependent processManager
	processManager*     m_processManager;
	//! application dependent selectionManager
	selectionManager* m_selectionManager;
};

#endif
